unit oki6295;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}
     math,dialogs,timer_engine,sysutils,sound_engine;

const
  MAX_SAMPLE_CHUNK=10000;
  OKIM6295_VOICES=4;
  OKIM6295_PIN7_LOW=0;
  OKIM6295_PIN7_HIGH=1;
  // step size index shift table */
  index_shift:array[0..7] of integer =(-1, -1, -1, -1, 2, 4, 6, 8 );
  // volume lookup table. The manual lists only 9 steps, ~3dB per step. Given the dB values,
  // that seems to map to a 5-bit volume control. Any volume parameter beyond the 9th index
  // results in silent playback.
  volume_table:array[0..15] of integer =(
	$20,	//   0 dB
	$16,	//  -3.2 dB
	$10,	//  -6.0 dB
	$0b,	//  -9.2 dB
	$08,	// -12.0 dB
	$06,	// -14.5 dB
	$04,	// -18.0 dB
	$03,	// -20.5 dB
	$02,	// -24.0 dB
	$00,
	$00,
	$00,
	$00,
	$00,
	$00,
	$00);

type
      adpcm_state=record
        signal:integer;
        step:integer;
      end;

      // struct describing a single playing ADPCM voice */
      ADPCMVoice=record
	        playing:boolean;			//if we are actively playing */
	        base_offset:dword;		// pointer to the base memory location */
	        sample:dword;			// current sample number */
	        count:dword;			// total samples to play */
          adpcm:adpcm_state; // current ADPCM state */
	        volume:dword;			// output volume
      end;

      tokim6295=record
	        voice:array[0..OKIM6295_VOICES] of ADPCMVoice;
          command:integer;
	        bank_installed:boolean;
	        bank_offs:integer;
	        master_clock:dword;	// master clock frequency */
          rom:pbyte;
          tsample:byte;
          snd_clock,clock:dword;
          amp,ntimer:byte;
          out_:integer;
      end;
      pokim6295=^tokim6295;
var
    //lookup table for the precomputed difference */
    diff_lookup:array[0..(49*16)-1] of double;
    oki_6295:array[0..1] of pokim6295;

function okim6295_r(num:byte):byte;
procedure okim6295_w(num,valor:byte);
procedure init_oki6295(num:byte;clock:dword;pin7,nsnd_cpu:byte;snd_clock:dword;amp:byte);
procedure close_oki6295(num:byte);
procedure reset_oki6295(num:byte);
procedure okim6295_update(num:byte);
procedure internal_update_oki6295_0;
procedure internal_update_oki6295_1;
procedure final_update_oki6295_0;
procedure final_update_oki6295_1;
procedure okim6295_change_pin7(num,pin7:byte);

implementation

procedure compute_tables;inline;
const
	// nibble to bit map */
	nbl2bit:array[0..15,0..3] of integer=(
		( 1, 0, 0, 0), ( 1, 0, 0, 1), ( 1, 0, 1, 0), ( 1, 0, 1, 1),
		( 1, 1, 0, 0), ( 1, 1, 0, 1), ( 1, 1, 1, 0), ( 1, 1, 1, 1),
		(-1, 0, 0, 0), (-1, 0, 0, 1), (-1, 0, 1, 0), (-1, 0, 1, 1),
		(-1, 1, 0, 0), (-1, 1, 0, 1), (-1, 1, 1, 0), (-1, 1, 1, 1));
var
	step,nib,stepval:integer;
begin
	// loop over all possible steps */
	for step:=0 to 48 do begin
		// compute the step value */
		stepval:= floor(16.0*power(11.0/10.0,step));
		// loop over all nibbles and compute the difference */
		for nib:=0 to 15 do begin
			diff_lookup[step*16 + nib]:=nbl2bit[nib][0]*(stepval*nbl2bit[nib][1]+
				 stepval/2*nbl2bit[nib][2]+
				 stepval/4*nbl2bit[nib][3]+
				 stepval/8);
		end;
  end;
end;

procedure reset_adpcm(num,num_voice:byte);inline;
var
  chip:pokim6295;
begin
  chip:=oki_6295[num];
	// reset the signal/step */
	chip.voice[num_voice].adpcm.signal:=-2;
	chip.voice[num_voice].adpcm.step:=0;
end;

function clock_adpcm(num,num_voice,nibble:byte):integer;inline;
var
  chip:pokim6295;
begin
  chip:=oki_6295[num];
	chip.voice[num_voice].adpcm.signal:=chip.voice[num_voice].adpcm.signal+round(diff_lookup[chip.voice[num_voice].adpcm.step*16+(nibble and 15)]);
	// clamp to the maximum */
	if (chip.voice[num_voice].adpcm.signal >2047) then chip.voice[num_voice].adpcm.signal:=2047
	  else if (chip.voice[num_voice].adpcm.signal<-2048) then chip.voice[num_voice].adpcm.signal:=-2048;
	// adjust the step size and clamp */
	chip.voice[num_voice].adpcm.step:=chip.voice[num_voice].adpcm.step+index_shift[nibble and 7];
	if (chip.voice[num_voice].adpcm.step>48) then chip.voice[num_voice].adpcm.step:=48
	  else if (chip.voice[num_voice].adpcm.step<0) then chip.voice[num_voice].adpcm.step:=0;
	// return the signal */
	clock_adpcm:=chip.voice[num_voice].adpcm.signal;
end;

function generate_adpcm(num,num_voice:byte):integer;inline;
var
  chip:pokim6295;
  nibble:byte;
  base,sample,count:integer;
  ptemp:pbyte;
begin
  chip:=oki_6295[num];
  base:=chip.voice[num_voice].base_offset;
  sample:=chip.voice[num_voice].sample;
  count:=chip.voice[num_voice].count;
  // compute the new amplitude and update the current step */
  ptemp:=chip.rom;
  inc(ptemp,base+(sample shr 1));
  nibble:=ptemp^;
  if (sample and 1)=0 then nibble:=nibble shr 4
    else nibble:=nibble and $f;//(((sample and 1) shl 2) xor 4);
  // output to the buffer, scaling by the volume */
  //signal in range -2048..2047, volume in range 2..32 => signal * volume / 2 in range -32768..32767 */
  generate_adpcm:=((clock_adpcm(num,num_voice,nibble)*(chip.voice[num_voice].volume shr 1)))*chip.amp;
  // next! */
  sample:=sample+1;
  if (sample>=count) then chip.voice[num_voice].playing:=false;
  // update the parameters */
  chip.voice[num_voice].sample:=sample;
end;

procedure reset_oki6295(num:byte);
var
  chip:pokim6295;
  f:byte;
begin
  chip:=oki_6295[num];
  // initialize the voices */
  chip.command:=-1;
  chip.out_:=0;
  chip.bank_offs:=0;
	for f:=0 to OKIM6295_VOICES do begin
		reset_adpcm(num,f);
    chip.voice[f].playing:=false;
    chip.voice[f].count:=0;
    chip.voice[f].base_offset:=0;
    chip.voice[f].sample:=0;
    chip.voice[f].volume:=0;
	end;
end;

procedure close_oki6295(num:byte);
begin
if oki_6295[num]<>nil then begin
  freemem(oki_6295[num].rom);
  oki_6295[num].rom:=nil;
  freemem(oki_6295[num]);
  oki_6295[num]:=nil;
end;
end;

procedure okim6295_change_pin7(num,pin7:byte);
var
  divisor:byte;
  chip:pokim6295;
begin
  chip:=oki_6295[num];
  if pin7=OKIM6295_PIN7_HIGH then divisor:=132
    else divisor:=165;
  timer[chip.ntimer].time_final:=chip.snd_clock/(chip.clock/divisor);
end;

procedure init_oki6295(num:byte;clock:dword;pin7,nsnd_cpu:byte;snd_clock:dword;amp:byte);
var
  chip:pokim6295;
begin
  if oki_6295[num]=nil then begin
    getmem(oki_6295[num],sizeof(tokim6295));
    fillchar(oki_6295[num]^,sizeof(tokim6295),0);
  end;
  chip:=oki_6295[num];
  getmem(chip.rom,$40000);
	compute_tables;
	chip.bank_installed:=false;
	chip.master_clock:=clock;
  chip.tsample:=init_channel;
  chip.amp:=amp;
  chip.snd_clock:=snd_clock;
  chip.clock:=clock;
  case num of
      0:begin
          chip.ntimer:=init_timer(nsnd_cpu,1,internal_update_oki6295_0,true);
          init_timer(nsnd_cpu,snd_clock/44100,final_update_oki6295_0,true);
        end;
      1:begin
          chip.ntimer:=init_timer(nsnd_cpu,1,internal_update_oki6295_1,true);
          init_timer(nsnd_cpu,snd_clock/44100,final_update_oki6295_1,true);
        end;
  end;
  okim6295_change_pin7(num,pin7);
	// initialize the voices */
  reset_oki6295(num);
end;

function okim6295_r(num:byte):byte;
var
  f,res:byte;
begin
	res:=$f0;	// naname expects bits 4-7 to be 1 */
	// set the bit to 1 if something is playing on a given channel */
	for f:=0 to OKIM6295_VOICES do
    if oki_6295[num].voice[f].playing then res:=res or (1 shl f);
	okim6295_r:=res;
end;

procedure okim6295_w(num,valor:byte);
var
  chip:pokim6295;
  base:word;
  start,stop:dword;
  ptemp:pbyte;
  temp,i:byte;
begin
  chip:=oki_6295[num];
	// if a command is pending, process the second half */
	if (chip.command<>-1) then begin
		temp:=valor shr 4;
		// the manual explicitly says that it's not possible to start multiple voices at the same time */
		//if ((temp<>0) and (temp<>1) and (temp<>2) and (temp<>4) and (temp<>8)) then
		//	MessageDlg('OKI 6295 - Oppps! Inicia mas de un canal a la vez!', mtInformation,[mbOk], 0);
		// determine which voice(s) (voice is set by a 1 bit in the upper 4 bits of the second byte) */
		for i:=0 to OKIM6295_VOICES do begin
			if (temp and 1)<>0 then begin
				// determine the start/stop positions */
				base:=chip.command*8;
        ptemp:=chip.rom;
        inc(ptemp,base);
				start:=ptemp^ shl 16;
        inc(ptemp);
				start:=start+(ptemp^ shl 8);
        inc(ptemp);
				start:=(start+ptemp^) and $3ffff;
        inc(ptemp);
				stop:=ptemp^ shl 16;
        inc(ptemp);
				stop:=stop+(ptemp^ shl 8);
        inc(ptemp);
				stop:=(stop+ptemp^) and $3ffff;
				// set up the voice to play this sample */
				if (start<stop) then begin
					if not(chip.voice[i].playing) then begin // fixes Got-cha and Steel Force */
						chip.voice[i].playing:=true;
						chip.voice[i].base_offset:=start;
						chip.voice[i].sample:=0;
						chip.voice[i].count:=2*(stop-start+1);
						// also reset the ADPCM parameters */
						reset_adpcm(num,i);
						chip.voice[i].volume:=volume_table[valor and $0f];
					end	else begin
						//logerror("OKIM6295:'%s' requested to play sample %02x on non-stopped voice\n",device->tag(),info->command);
					end;
				end else begin 	// invalid samples go here */
					//logerror("OKIM6295:'%s' requested to play invalid sample %02x\n",device->tag(),info->command);
					chip.voice[i].playing:=false;
				end;
			end;
      temp:=temp shr 1;
		end;  //del for
		// reset the command */
		chip.command:=-1;
  end	else begin // if this is the start of a command, remember the sample number for next time */
    if (valor and $80)<>0 then begin
		      chip.command:=valor and $7f;
    end else begin  // otherwise, see if this is a silence command */
                temp:=valor shr 3;
                //determine which voice(s) (voice is set by a 1 bit in bits 3-6 of the command */
                for i:=0 to OKIM6295_VOICES do begin
                  if (temp and 1)<>0 then chip.voice[i].playing:=false;
                  temp:=temp shr 1;
                end;
          end;
  end;
end;

procedure okim6295_stream_update(num:byte);
var
  f:byte;
  chip:pokim6295;
begin
  chip:=oki_6295[num];
  chip.out_:=0;
	for f:=0 to OKIM6295_VOICES do
    if chip.voice[f].playing then chip.out_:=chip.out_+generate_adpcm(num,f);
  if chip.out_<-32767 then chip.out_:=-32767
    else if chip.out_>32767 then chip.out_:=32767;
end;

procedure internal_update_oki6295_0;
begin
  okim6295_stream_update(0);
end;

procedure internal_update_oki6295_1;
begin
  okim6295_stream_update(1);
end;

procedure okim6295_update(num:byte);
var
  chip:pokim6295;
begin
  chip:=oki_6295[num];
  tsample[chip.tsample,sound_status.posicion_sonido]:=chip.out_;
  if sound_status.stereo then tsample[chip.tsample,sound_status.posicion_sonido+1]:=chip.out_;
end;

procedure final_update_oki6295_0;
begin
  okim6295_update(0);
end;

procedure final_update_oki6295_1;
begin
  okim6295_update(1);
end;

end.
